/* 2018-08-15: File added and changed by Sony Corporation */
/*
 *  tegra18x_wdt.c: watchdog driver for NVIDIA tegra internal watchdog
 *
 *  Copyright 2018 Sony Corporation
 *  Based on drivers/watchdog/tegra_wdt_t18x.c in
 *  https://nv-tegra.nvidia.com/gitweb/?p=linux-t18x.git;a=summary
 *
 *  Copyright (c) 2012-2016, NVIDIA CORPORATION. All rights reserved.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <common.h>
#include <asm/io.h>
#include <watchdog.h>
#include <fdtdec.h>

DECLARE_GLOBAL_DATA_PTR;

#define TOP_TKE_TKEIE_BASE          0x100
#define TOP_TKE_TKEIE(i)            (0x100 + 4 * (i))
#define TOP_TKE_TKEIE_WDT_MASK(i)   (1 << (16 + 4 * (i)))

#define TOP_TKE_TMR_PTV             0
#define TOP_TKE_TMR_EN              BIT(31)
#define TOP_TKE_TMR_PERIODIC        BIT(30)
#define TOP_TKE_TMR_PTV_MASK        0xFFFFFFF

#define TOP_TKE_TMR_PCR             0x4
#define TOP_TKE_TMR_PCR_INTR        BIT(30)
#define WDT_CFG                     (0)
#define WDT_CFG_PERIOD              BIT(4)
#define WDT_CFG_INT_EN              BIT(12)
#define WDT_CFG_FINT_EN             BIT(13)
#define WDT_CFG_REMOTE_INT_EN       BIT(14)
#define WDT_CFG_DBG_RST_EN          BIT(15)
#define WDT_CFG_SYS_PORST_EN        BIT(16)
#define WDT_CFG_ERR_THRESHOLD       (7 << 20)
#define WDT_STATUS                  (0x4)
#define WDT_INTR_STAT               BIT(1)
#define WDT_CMD                     (0x8)
#define WDT_CMD_START_COUNTER       BIT(0)
#define WDT_CMD_DISABLE_COUNTER     BIT(1)
#define WDT_UNLOCK                  (0xC)
#define WDT_UNLOCK_PATTERN          (0xC45A << 0)
#define WDT_SKIP                    (0x10)
#define WDT_SKIP_VAL(i, val)        (((val) & 0xf) << (4 * (i)))


struct tegra_wdt_t18x {
	void *wdt_source;
	void *wdt_timer;
	void *wdt_tke;
	u32  timeout;
	u32  config;
	int  expiry_count;
	int  active_count;
	int  index;
	bool config_locked;
	bool enabled;
	bool feed;
	u32  timeout_max;
	u32  timeout_min;
	u32  reset_flags;
};

#define USEC_PER_MSEC     1000L
#define MSEC_PER_SEC      1000L
#define USEC_PER_SEC      (USEC_PER_MSEC * MSEC_PER_SEC)

static struct tegra_wdt_t18x wdt = {
	.feed         = true,
};
static struct tegra_wdt_t18x *twdt_t18x = &wdt;

static inline int tegra_wdt_t18x_skip(struct tegra_wdt_t18x *twdt_t18x)
{
	u32 val = 0;
	int skip_count = 0;

	/* Skip the 4th expiry if debug reset is disabled */
	bool remote_skip = !(twdt_t18x->config & WDT_CFG_REMOTE_INT_EN);
	bool dbg_skip = !(twdt_t18x->config & WDT_CFG_DBG_RST_EN);

	if (remote_skip) {
		if (dbg_skip) {
			val |= WDT_SKIP_VAL(2, 2);
			skip_count += 2;
		} else {
			val |= WDT_SKIP_VAL(2, 1);
			skip_count += 1;
		}
	} else {
		if (dbg_skip) {
			val |= WDT_SKIP_VAL(3, 1);
		}
	}

	debug("%s: remote_skip(%d),  dbg_skip(%d) , skip_count(%d)\n",
			__func__, remote_skip, dbg_skip, skip_count);

	if (val)
		writel(val, twdt_t18x->wdt_source + WDT_SKIP);

	return skip_count;
}

#ifdef DEBUG
void debug_writel(u32 val, void *addr)
{
	printf("W: %p: %08x\n", addr, val);
	writel(val, addr);
}
#undef writel
#define writel(v, a) debug_writel(v, a)
#endif
static void tegra_wdt_t18x_disable(struct tegra_wdt_t18x *twdt_t18x)
{
	writel(WDT_UNLOCK_PATTERN, twdt_t18x->wdt_source + WDT_UNLOCK);
	writel(WDT_CMD_DISABLE_COUNTER, twdt_t18x->wdt_source + WDT_CMD);

	writel(0, twdt_t18x->wdt_timer + TOP_TKE_TMR_PTV);
}

static void tegra_wdt_t18x_check_timeout(struct tegra_wdt_t18x *twdt_t18x)
{
	if (twdt_t18x->timeout < twdt_t18x->timeout_min ||
		twdt_t18x->timeout > twdt_t18x->timeout_max) {
		twdt_t18x->timeout = CONFIG_WATCHDOG_TIMEOUT_MSECS * USEC_PER_MSEC;
		puts("timeout value out of range, use the default setting\n");
	}
}

static void tegra_wdt_t18x_enable(struct tegra_wdt_t18x *twdt_t18x)
{
	u32 val;

	/* Update skip configuration and active expiry count */
	twdt_t18x->active_count = twdt_t18x->expiry_count -
					 tegra_wdt_t18x_skip(twdt_t18x);
	if (twdt_t18x->active_count < 1)
		twdt_t18x->active_count = 1;

	/*
	 * Some commands don't call WATCHDOG_RESET(), and most of them spend
	 * less than 0.5s
	 */
	twdt_t18x->timeout_min = 500 * USEC_PER_MSEC;
	twdt_t18x->timeout_max = TOP_TKE_TMR_PTV_MASK * twdt_t18x->active_count;

	tegra_wdt_t18x_check_timeout(twdt_t18x);

	writel(TOP_TKE_TMR_PCR_INTR, twdt_t18x->wdt_timer + TOP_TKE_TMR_PCR);
	/*
	 * The timeout needs to be divided by active expiry count here so as
	 * to keep the ultimate watchdog reset timeout the same as the program
	 * timeout requested by application. The program timeout should make
	 * sure WDT FIQ will never be asserted in a valid use case.
	 */
	val = twdt_t18x->timeout / twdt_t18x->active_count;
	val |= (TOP_TKE_TMR_EN | TOP_TKE_TMR_PERIODIC);
	writel(val, twdt_t18x->wdt_timer + TOP_TKE_TMR_PTV);

	if (!twdt_t18x->config_locked)
		writel(twdt_t18x->config, twdt_t18x->wdt_source + WDT_CFG);
	writel(WDT_CMD_START_COUNTER, twdt_t18x->wdt_source + WDT_CMD);
}

static void tegra_wdt_t18x_init(struct tegra_wdt_t18x *twdt_t18x)
{
	twdt_t18x->config_locked = !!(twdt_t18x->config & WDT_CFG_INT_EN);

	if (!twdt_t18x->config_locked) {
		/* Configure timer source and period */
		twdt_t18x->config = (((phys_addr_t)twdt_t18x->wdt_source >> 16) & (0xf)) - 2;
		twdt_t18x->config |= WDT_CFG_PERIOD;

		/* Enable local interrupt for WDT petting */
		twdt_t18x->config |= WDT_CFG_INT_EN;

		/*
		 * 'ErrorThreshold' field @ TKE_TOP_WDT1_WDTCR_0 decides the
		 * indication to HSM. The WDT logic asserts an error signal to
		 * HSM when ExpirationLevel >= ErrorThreshold. Retain the POR
		 * value to avoid nuisance trigger to HSM.
		 */
		twdt_t18x->config |= WDT_CFG_ERR_THRESHOLD;

		/* Enable local FIQ for debug dump */
		twdt_t18x->config |= WDT_CFG_FINT_EN;

		twdt_t18x->config |= twdt_t18x->reset_flags;
	}

	tegra_wdt_t18x_disable(twdt_t18x);
	writel(TOP_TKE_TMR_PCR_INTR, twdt_t18x->wdt_timer + TOP_TKE_TMR_PCR);

	tegra_wdt_t18x_enable(twdt_t18x);

	twdt_t18x->enabled = true;

	hw_watchdog_reset();
}

#ifdef DEBUG
void tegra_wdt_t18x_dump(struct tegra_wdt_t18x *twdt_t18x)
{
	printf("wdt_source = %p\n", twdt_t18x->wdt_source);
	printf("wdt_timer = %p\n", twdt_t18x->wdt_timer);
	printf("wdt_tke = %p\n", twdt_t18x->wdt_tke);
	printf("expiry_count = %d\n", twdt_t18x->expiry_count);
	printf("timeout = %uus\n", twdt_t18x->timeout);
	printf("reset_flags=%x\n", twdt_t18x->reset_flags);
}
#endif

void hw_watchdog_init(void)
{
	int node, ret;
	struct fdt_resource res;
	const void *fdt = gd->fdt_blob;

	if (NULL == fdt) {
		printf("Error: no device tree found\n");
		return;
	}

	node = fdt_node_offset_by_compatible(fdt, -1, "nvidia,tegra-wdt-t18x");
	if (node < 0) {
		printf("Error: no device found\n");
		return;
	}

	if (!fdtdec_get_is_enabled(fdt, node))
		return;

	/* TOP_TKE:TOP_TKE_WDT:WDT0 module */
	ret = fdt_get_resource(fdt, node, "reg", 0, &res);
	if (ret < 0)
		return;
	twdt_t18x->wdt_source = (void *)res.start;

	/* TOP_TKE:TOP_TKE_TMR:TMR7 module */
	ret = fdt_get_resource(fdt, node, "reg", 1, &res);
	if (ret < 0)
		return;
	twdt_t18x->wdt_timer = (void *)res.start;

	/* TOP_TKE:TMR_SHARED module */
	ret = fdt_get_resource(fdt, node, "reg", 2, &res);
	if (ret < 0)
		return;
	twdt_t18x->wdt_tke = (void *)res.start;

	twdt_t18x->expiry_count = fdtdec_get_uint(fdt, node,
			"nvidia,expiry-count", 5);

	/* If CONFIG_HW_WATCHDOG=y or user set timeout to 0, get from dtb */
	if (twdt_t18x->timeout == 0) {
		twdt_t18x->timeout = fdtdec_get_uint(fdt, node, "timeout-sec", 0);
		if (twdt_t18x->timeout) {
			printf("Watchdog timeout is set to %lums by device tree\n",
					twdt_t18x->timeout * MSEC_PER_SEC);
			twdt_t18x->timeout *= USEC_PER_SEC;
		}
	}

	twdt_t18x->config = readl(twdt_t18x->wdt_source + WDT_CFG);
	if (fdtdec_get_bool(fdt, node, "nvidia,disable-remote-interrupt")) {
		twdt_t18x->config &= ~WDT_CFG_REMOTE_INT_EN;
	}

	/* Debug and POR reset events should be enabled by default. */
	twdt_t18x->reset_flags = WDT_CFG_DBG_RST_EN | WDT_CFG_SYS_PORST_EN;
	if (fdtdec_get_bool(fdt, node, "nvidia,disable-debug-reset"))
		twdt_t18x->reset_flags &= ~WDT_CFG_DBG_RST_EN;
	if (fdtdec_get_bool(fdt, node, "nvidia,disable-por-reset"))
		twdt_t18x->reset_flags &= ~WDT_CFG_SYS_PORST_EN;

#ifdef DEBUG
	tegra_wdt_t18x_dump(twdt_t18x);
#endif

	tegra_wdt_t18x_init(twdt_t18x);
}

void hw_watchdog_reset(void)
{
	u32 val;

	if (!twdt_t18x->enabled || !twdt_t18x->feed)
		return;

	tegra_wdt_t18x_check_timeout(twdt_t18x);

	/* Disable timer, load the timeout value and restart. */
	writel(WDT_UNLOCK_PATTERN, twdt_t18x->wdt_source + WDT_UNLOCK);
	writel(WDT_CMD_DISABLE_COUNTER, twdt_t18x->wdt_source + WDT_CMD);

	writel(TOP_TKE_TMR_PCR_INTR, twdt_t18x->wdt_timer + TOP_TKE_TMR_PCR);

	val = twdt_t18x->timeout / twdt_t18x->active_count;
	val |= (TOP_TKE_TMR_EN | TOP_TKE_TMR_PERIODIC);
	writel(val, twdt_t18x->wdt_timer + TOP_TKE_TMR_PTV);

	writel(WDT_CMD_START_COUNTER, twdt_t18x->wdt_source + WDT_CMD);
}

#ifdef CONFIG_CMD_WDT
void hw_watchdog_feed_enable(bool yesno)
{
	if (!twdt_t18x->enabled)
		printf("Warning: watchdog is not enabled\n");

	twdt_t18x->feed = yesno;
}

u32 hw_watchdog_start(u32 timeout_ms)
{
	twdt_t18x->timeout = timeout_ms * USEC_PER_MSEC;

	hw_watchdog_init();

	return twdt_t18x->timeout / USEC_PER_MSEC;
}
#endif
